searchNodes=[{"ref":"Norm.html","title":"Norm","type":"module","doc":"Norm is a system for specifying the structure of data. It can be used for validation and for generation of data. Norm does not provide any set of predicates and instead allows you to re-use any of your existing validations. import Norm conform!(123, spec(is_integer() and &amp;(&amp; &gt; 0))) =&gt; 123 conform!(-50, spec(is_integer() and &amp;(&amp;1 &gt; 0))) ** (Norm.MismatchError) val: -50 spec: &amp;(&amp;1 &gt; 0) (norm) lib/norm.ex:44: Norm.conform!/2 user_schema = schema(%{ user: schema(%{ name: spec(is_binary()), age: spec(is_integer() and &amp;(&amp;1 &gt; 0)) }) }) conform!(%{user: %{name: &quot;chris&quot;, age: 30}}, user_schema) =&gt; %{user: %{name: &quot;chris&quot;, age: 30}} user_schema |&gt; gen() |&gt; Enum.take(3) =&gt; [ %{user: %{age: 0, name: &quot;&quot;}}, %{user: %{age: 2, name: &quot;x&quot;}}, %{user: %{age: -2, name: &quot;&quot;}} ] Validation and conforming values Norm validates data by &quot;conforming&quot; the value to a specification. If the values don&#39;t conform then a list of errors is returned. There are 2 functions provided for this conform/2 and conform!/2. If you need to return a list of well defined errors then you should use conform/2. Otherwise conform!/2 is generally more useful. The input data is always passed as the 1st argument to conform so that calls to conform are easily chainable. Predicates and specs Norm does not provide a special set of predicates and instead allows you to convert any predicate into a spec with the spec/1 macro. Predicates can be composed together using the and and or keywords. You can also use anonymous functions to create specs. spec(is_binary()) spec(is_integer() and &amp;(&amp;1 &gt; 0)) spec(is_binary() and fn str -&gt; String.length(str) &gt; 0 end) The data is always passed as the first argument to your predicate so you can use predicates with multiple values like so: def greater?(x, y), do: x &gt; y conform!(10, spec(greater?(5))) =&gt; 10 conform!(3, spec(greater?(5))) ** (Norm.MismatchError) val: 3 spec: greater?(5) (norm) lib/norm.ex:44: Norm.conform!/2 Tuples and atoms Atoms and tuples can be matched without needing to wrap them in a function. :some_atom = conform!(:some_atom, :atom) {1, &quot;hello&quot;} = conform!({1, &quot;hello&quot;}, {spec(is_integer()), spec(is_binary())}) conform!({1, 2}, {:one, :two}) ** (Norm.MismatchError) val: 1 in: 0 fails: is not an atom. val: 2 in: 1 fails: is not an atom. Because Norm supports matching on bare tuples we can easily validate functions that return {:ok, term()} and {:error, term()} tuples. # if User.get_name/1 succeeds it returns {:ok, binary()} result = User.get_name(123) {:ok, name} = conform!(result, {:ok, spec(is_binary())}) These specifications can be combined with one_of/1 to create union types. result_spec = one_of([ {:ok, spec(is_binary())}, {:error, spec(fn _ -&gt; true end)}, ]) {:ok, &quot;alice&quot;} = conform!(User.get_name(123), result_spec) {:error, &quot;user does not exist&quot;} = conform!(User.get_name(-42), result_spec) Schemas Norm provides a schema/1 function for specifying maps and structs: user_schema = schema(%{ user: schema(%{ name: spec(is_binary()), age: spec(is_integer()), }) }) conform!(%{user: %{name: &quot;chris&quot;, age: 31}}, user_schema) =&gt; %{user: %{name: &quot;chris&quot;, age: 31}} conform!(%{user: %{name: &quot;chris&quot;, age: -31}}, user_schema) ** (Norm.MismatchError) in: :user/:age val: -31 fails: &amp;(&amp;1 &gt; 0) (norm) lib/norm.ex:44: Norm.conform!/2 You can also create specs from structs: defmodule User do defstruct [:name, :age] def s, do: schema(%__MODULE__{ name: spec(is_binary()), age: spec(is_integer()) }) end This will ensure that the input is a User struct with the key that match the given specification. Its convention to provide a s() function in the module that defines the struct so that schema&#39;s can be shared throughout your system. You don&#39;t need to provide specs for all the keys in your struct. Only the specced keys will be conformed. The remaining keys will be checked for presence. defmodule User do defstruct [:name, :age] end conform!(%User{name: &quot;chris&quot;}, schema(%User{})) =&gt; %User{name: &quot;chris&quot;, age: nil} Key semantics Atom and string keys are matched explicitly and there is no casting that occurs when conforming values. If you need to match on string keys you should specify your schema with string keys. Schema&#39;s accomodate growth by disregarding any unspecified keys in the input map. This allows callers to start sending new data over time without coordination with the consuming function. Selections You may have noticed that there&#39;s no way to specify optional keys in a schema. This may seem like an oversight but its actually an intentional design decision. Whether a key should be present in a schema is determined by the call site and not by the schema itself. For instance think about the assigns in a plug conn. When are the assigns optional? It depends on where you are in the pipeline. Schema&#39;s also force all keys to match at all times. This is generally useful as it limits your ability to introduce errors. But it also limits schema growth and turns changes that should be non-breaking into breaking changes. In order to support both of these scenarios Norm provides the selection/2 function. selection/2 allows you to specify exactly the keys you require from a schema at the place where you require them. user_schema = schema(%{ user: schema(%{ name: spec(is_binary()), age: spec(is_integer()), }) }) just_age = selection(user_schema, [user: [:age]]) conform!(%{user: %{name: &quot;chris&quot;, age: 31}}, just_age) =&gt; %{user: %{age: 31}} # Selection also disregards unspecified keys conform!(%{user: %{name: &quot;chris&quot;, age: 31, unspecified: nil}, other_stuff: :foo}, just_age) =&gt; %{user: %{age: 31}} Patterns Norm provides a way to specify alternative specs using the alt/1 function. This is useful when you need to support multiple schema&#39;s or multiple alternative specs. create_event = schema(%{type: spec(&amp;(&amp;1 == :create))}) update_event = schema(%{type: spec(&amp;(&amp;1 == :update))}) event = alt(create: create_event, update: update_event) conform!(%{type: :create}, event) =&gt; {:create, %{type: :create}} conform!(%{type: :update}, event) =&gt; {:update, %{type: :update}} conform!(%{type: :delete}, event) ** (Norm.MismatchError) val: :delete in: :create/:type fails: &amp;(&amp;1 == :create) val: :delete in: :update/:type fails: &amp;(&amp;1 == :update) Generators Along with validating that data conforms to a given specification, Norm can also use specificiations to generate examples of good data. These examples can then be used for property based testing, local development, seeding databases, or any other use case. user_schema = schema(%{ user: schema(%{ name: spec(is_binary()), age: spec(is_integer() and &amp;(&amp;1 &gt; 0)) }) }) conform!(%{user: %{name: &quot;chris&quot;, age: 30}}, user_schema) =&gt; %{user: %{name: &quot;chris&quot;, age: 30}} user_schema |&gt; gen() |&gt; Enum.take(3) =&gt; [ %{user: %{age: 0, name: &quot;&quot;}}, %{user: %{age: 2, name: &quot;x&quot;}}, %{user: %{age: -2, name: &quot;&quot;}} ] Under the hood Norm uses StreamData for its data generation. This means you can use your specs in tests like so: input_data = schema(%{&quot;user&quot; =&gt; schema(%{&quot;name&quot; =&gt; spec(is_binary())})}) property &quot;users can update names&quot; do check all input &lt;- gen(input_data) do assert :ok == update_user(input) end end Built in generators Norm will try to infer the generator to use from the predicate defined in spec. It looks specifically for the guard clauses used for primitive types in elixir. Not all of the built in guard clauses are supported yet. PRs are very welcome ;). Guiding generators You may have specs like spec(fn x -&gt; rem(x, 2) == 0 end) which check to see that an integer is even or not. This generator expects integer values but there&#39;s no way for Norm to determine this. If you try to create a generator from this spec you&#39;ll get an error: gen(spec(fn x -&gt; rem(x, 2) == 0 end)) ** (Norm.GeneratorError) Unable to create a generator for: fn x -&gt; rem(x, 2) == 0 end (norm) lib/norm.ex:76: Norm.gen/1 You can guide Norm to the right generator by specifying a guard clause as the first predicate in a spec. If Norm can find the right generator then it will use any other predicates as filters in the generator. Enum.take(gen(spec(is_integer() and fn x -&gt; rem(x, 2) == 0 end)), 5) [0, -2, 2, 0, 4] But its also possible to create filters that are too specific such as this: gen(spec(is_binary() and &amp;(&amp;1 =~ ~r/foobarbaz/))) Norm can determine the generators to use however its incredibly unlikely that Norm will be able to generate data that matches the filter. After 25 consecutive unseccessful attempts to generate a good value Norm (StreamData under the hood) will return an error. In these scenarios we can create a custom generator. Overriding generators You&#39;ll often need to guide your generators into the interesting parts of the state space so that you can easily find bugs. That means you&#39;ll want to tweak and control your generators. Norm provides an escape hatch for creating your own generators with the with_gen/2 function: age = spec(is_integer() and &amp;(&amp;1 &gt;= 0)) reasonable_ages = with_gen(age, StreamData.integer(0..105)) Because gen/1 returns a StreamData generator you can compose your generators with other StreamData functions: age = spec(is_integer() and &amp;(&amp;1 &gt;= 0)) StreamData.frequencies([ {3, gen(age)}, {1, StreamData.binary()}, ]) gen(age) |&gt; StreamData.map(&amp;Integer.to_string/1) |&gt; Enum.take(5) [&quot;1&quot;, &quot;1&quot;, &quot;3&quot;, &quot;4&quot;, &quot;1&quot;] This allows you to compose generators however you need to while keeping your generation co-located with the specification of the data."},{"ref":"Norm.html#alt/1","title":"Norm.alt/1","type":"function","doc":"Chooses between alternative predicates or patterns. The patterns must be tagged with an atom. When conforming data to this specification the data is returned as a tuple with the tag. Examples iex&gt; conform!(&quot;foo&quot;, alt(s: spec(is_binary()), a: spec(is_atom()))) {:s, &quot;foo&quot;} iex&gt; conform!(:foo, alt(s: spec(is_binary()), a: spec(is_atom()))) {:a, :foo} iex&gt; conform!(123, alt(num: spec(is_integer()), str: spec(is_binary()))) {:num, 123} iex&gt; conform!(&quot;foo&quot;, alt(num: spec(is_integer()), str: spec(is_binary()))) {:str, &quot;foo&quot;} iex&gt; conform(true, alt(num: spec(is_integer()), str: spec(is_binary()))) {:error, [%{spec: &quot;is_integer()&quot;, input: true, path: [:num]}, %{spec: &quot;is_binary()&quot;, input: true, path: [:str]}]}"},{"ref":"Norm.html#coll_of/2","title":"Norm.coll_of/2","type":"function","doc":"Specifies a generic collection. Collections can be any enumerable type. Examples iex&gt; conform!([:a, :b, :c], coll_of(spec(is_atom()))) [:a, :b, :c]"},{"ref":"Norm.html#conform/2","title":"Norm.conform/2","type":"function","doc":"Verifies that the payload conforms to the specification. A &quot;success tuple&quot; is returned that contains either the conformed value or the error explanation. Examples: iex&gt; conform(42, spec(is_integer())) {:ok, 42} iex&gt; conform(42, spec(fn x -&gt; x == 42 end)) {:ok, 42} iex&gt; conform(42, spec(&amp;(&amp;1 &gt;= 0))) {:ok, 42} iex&gt; conform(42, spec(&amp;(&amp;1 &gt;= 100))) {:error, [%{spec: &quot;&amp;(&amp;1 &gt;= 100)&quot;, input: 42, path: []}]} iex&gt; conform(&quot;foo&quot;, spec(is_integer())) {:error, [%{spec: &quot;is_integer()&quot;, input: &quot;foo&quot;, path: []}]}"},{"ref":"Norm.html#conform!/2","title":"Norm.conform!/2","type":"function","doc":"Returns the conformed value or raises a mismatch error. Examples iex&gt; conform!(42, spec(is_integer())) 42 iex&gt; conform!(42, spec(is_binary())) ** (Norm.MismatchError) Could not conform input: val: 42 fails: is_binary()"},{"ref":"Norm.html#gen/1","title":"Norm.gen/1","type":"function","doc":"Creates a generator from a spec, schema, or selection. Examples iex&gt; gen(spec(is_integer())) |&gt; Enum.take(3) |&gt; Enum.all?(&amp;is_integer/1) true iex&gt; gen(spec(is_binary())) |&gt; Enum.take(3) |&gt; Enum.all?(&amp;is_binary/1) true iex&gt; gen(spec(&amp;(&amp;1 &gt; 0))) ** (Norm.GeneratorError) Unable to create a generator for: &amp;(&amp;1 &gt; 0)"},{"ref":"Norm.html#map_of/3","title":"Norm.map_of/3","type":"function","doc":"Specifies a map with a type of key and a type of value. Examples iex&gt; conform!(%{a: 1, b: 2, c: 3}, map_of(spec(is_atom()), spec(is_integer()))) %{a: 1, b: 2, c: 3}"},{"ref":"Norm.html#one_of/1","title":"Norm.one_of/1","type":"function","doc":"Chooses between a list of options. Unlike alt/1 the options don&#39;t need to be tagged. Specs are always tested in order and will short circuit if the data passes a validation. Examples iex&gt; conform!(&quot;chris&quot;, one_of([spec(is_binary()), :alice])) &quot;chris&quot; iex&gt; conform!(:alice, one_of([spec(is_binary()), :alice])) :alice"},{"ref":"Norm.html#schema/1","title":"Norm.schema/1","type":"function","doc":"Creates a re-usable schema. Schema&#39;s are open which means that all keys are optional and any non-specified keys are passed through without being conformed. If you need to mark keys as required instead of optional you can use selection. Examples iex&gt; valid?(%{}, schema(%{name: spec(is_binary())})) true iex&gt; valid?(%{name: &quot;Chris&quot;}, schema(%{name: spec(is_binary())})) true iex&gt; valid?(%{name: &quot;Chris&quot;, age: 31}, schema(%{name: spec(is_binary())})) true iex&gt; valid?(%{age: 31}, schema(%{name: spec(is_binary())})) true iex&gt; valid?(%{name: 123}, schema(%{name: spec(is_binary())})) false iex&gt; conform!(%{}, schema(%{name: spec(is_binary())})) %{} iex&gt; conform!(%{age: 31, name: &quot;chris&quot;}, schema(%{name: spec(is_binary())})) %{age: 31, name: &quot;chris&quot;} iex&gt; conform!(%{age: 31}, schema(%{name: spec(is_binary())})) %{age: 31} iex&gt; conform!(%{user: %{name: &quot;chris&quot;}}, schema(%{user: schema(%{name: spec(is_binary())})})) %{user: %{name: &quot;chris&quot;}}"},{"ref":"Norm.html#selection/2","title":"Norm.selection/2","type":"function","doc":"Selections can be used to mark keys on a schema as required. Any unspecified keys in the selection are still considered optional. Selections, like schemas, are open and allow unspecied keys to be passed through. If no selectors are provided then selection defaults to :all and recursively marks all keys in all nested schema&#39;s. If the schema includes internal selections these selections will not be overwritten. Examples iex&gt; valid?(%{name: &quot;chris&quot;}, selection(schema(%{name: spec(is_binary())}), [:name])) true iex&gt; valid?(%{}, selection(schema(%{name: spec(is_binary())}), [:name])) false iex&gt; valid?(%{user: %{name: &quot;chris&quot;}}, selection(schema(%{user: schema(%{name: spec(is_binary())})}), [user: [:name]])) true iex&gt; conform!(%{name: &quot;chris&quot;}, selection(schema(%{name: spec(is_binary())}), [:name])) %{name: &quot;chris&quot;} iex&gt; conform!(%{name: &quot;chris&quot;, age: 31}, selection(schema(%{name: spec(is_binary())}), [:name])) %{name: &quot;chris&quot;, age: 31} Require all keys iex&gt; valid?(%{user: %{name: &quot;chris&quot;}}, selection(schema(%{user: schema(%{name: spec(is_binary())})}))) true"},{"ref":"Norm.html#spec/1","title":"Norm.spec/1","type":"macro","doc":"Creates a new spec. Specs can be created from any existing predicates or anonymous functions. Specs must return a boolean value. Predicates can be arbitrarily composed using the and and or keywords. Examples: iex&gt; conform!(21, spec(is_integer())) 21 iex&gt; conform!(21, spec(is_integer() and &amp;(&amp;1 &gt;= 21))) 21 iex&gt; conform(&quot;21&quot;, spec(is_integer() and &amp;(&amp;1 &gt;= 21))) {:error, [%{spec: &quot;is_integer()&quot;, input: &quot;21&quot;, path: []}]} iex&gt; conform!(:foo, spec(is_atom() or is_binary())) :foo iex&gt; conform!(&quot;foo&quot;, spec(is_atom() or is_binary())) &quot;foo&quot; iex&gt; conform(21, spec(is_atom() or is_binary())) {:error, [%{spec: &quot;is_atom()&quot;, input: 21, path: []}, %{spec: &quot;is_binary()&quot;, input: 21, path: []}]}"},{"ref":"Norm.html#valid?/2","title":"Norm.valid?/2","type":"function","doc":"Checks if the value conforms to the spec and returns a boolean. Examples iex&gt; valid?(42, spec(is_integer())) true iex&gt; valid?(&quot;foo&quot;, spec(is_integer())) false"},{"ref":"Norm.html#with_gen/2","title":"Norm.with_gen/2","type":"function","doc":"Overwrites the default generator with a custom generator. The generator can be any valid StreamData generator. This means you can either use Norms built in gen/1 function or you can drop into StreamData directly. Examples iex&gt; Enum.take(gen(with_gen(spec(is_integer()), StreamData.constant(&quot;hello world&quot;))), 3) [&quot;hello world&quot;, &quot;hello world&quot;, &quot;hello world&quot;]"},{"ref":"Norm.Contract.html","title":"Norm.Contract","type":"module","doc":"Design by Contract with Norm. This module provides a @contract macro that can be used to define specs for arguments and the return value of a given function. To use contracts, call use Norm which also imports all Norm functions. Sometimes you may want to turn off contracts checking. For example, to skip contracts in production, set: config :norm, enable_contracts: Mix.env != :prod. Examples defmodule Colors do use Norm def rgb(), do: spec(is_integer() and &amp;(&amp;1 in 0..255)) def hex(), do: spec(is_binary() and &amp;String.starts_with?(&amp;1, &quot;#&quot;)) @contract rgb_to_hex(r :: rgb(), g :: rgb(), b :: rgb()) :: hex() def rgb_to_hex(r, g, b) do # ... end end"},{"ref":"Norm.GeneratorError.html","title":"Norm.GeneratorError","type":"exception","doc":""},{"ref":"Norm.GeneratorLibraryError.html","title":"Norm.GeneratorLibraryError","type":"exception","doc":""},{"ref":"Norm.MismatchError.html","title":"Norm.MismatchError","type":"exception","doc":""},{"ref":"Norm.SpecError.html","title":"Norm.SpecError","type":"exception","doc":""}]